import ipaddress
from abc import ABCMeta, abstractmethod
from typing import List, Tuple, Optional


class HostMeta(metaclass=ABCMeta):

    @property
    @abstractmethod
    def image_type(self):
        pass

    @property
    @abstractmethod
    def host_name(self):
        pass

    @property
    @abstractmethod
    def ipv4_address(self):
        pass

    @property
    @abstractmethod
    def ipv4_address_hex(self):
        pass

    @property
    @abstractmethod
    def aliases(self):
        pass

    @property
    @abstractmethod
    def entry(self):
        pass


class DefaultHost(HostMeta):

    def __init__(self, image_type):
        self._image_type = image_type

    @property
    def image_type(self):
        return self._image_type

    @property
    def host_name(self):
        return None

    @property
    def ipv4_address(self):
        return None

    @property
    def ipv4_address_hex(self):
        return "default"

    @property
    def aliases(self):
        return []

    @property
    def entry(self):
        return ""


class Host(HostMeta):

    def __init__(self, name_prefix: str, network: ipaddress.IPv4Network, address: ipaddress.IPv4Address, image_type: Optional[str] = None,
                 aliases: Optional[List[str]] = None, config: Optional[List[Tuple[str]]] = None, address_digits=2):
        self.host_name_format = lambda local_address: name_prefix + ("{:0" + str(address_digits) + "}").format(local_address)[-address_digits:]

        self._address = address
        self._network = network

        self._image_type = image_type
        self._aliases = tuple(aliases) if aliases is not None else ()
        self._config = tuple(config) if config is not None else ()

    @property
    def ipv4_address(self) -> str:
        return str(self._address)

    @property
    def reverse_pointer(self) -> str:
        return self._address.reverse_pointer

    @property
    def ipv4_address_hex(self) -> str:
        return "{:08X}".format(self._address._ip)

    @property
    def host_name(self):
        addr_mask = (1 << 32 - self._network.prefixlen) - 1
        return self.host_name_format(self._address._ip & addr_mask)

    @property
    def image_type(self) -> str:
        return self._image_type

    @property
    def aliases(self) -> Tuple[str]:
        return self._aliases

    @property
    def config(self):
        return self._config

    @property
    def entry(self) -> str:
        return "{}\t{}".format(self.ipv4_address, " ".join((self.host_name,) + self.aliases))

    def __str__(self):
        return str(self._address)


class HostGroup(object):

    def __init__(self, name_prefix: str, cidr_string: str):
        self.network = ipaddress.ip_network(cidr_string)
        self.name_prefix = name_prefix
        self._hosts = []

    def add_hosts(self, machine_offset: int, machine_count: int, aliases: Optional[List[str]] = None,
                  image_type: Optional[str] = None, config: Optional[Tuple[str]] = None) -> None:
        self._hosts += (Host(self.name_prefix, self.network, x, image_type, aliases, config) for i, x in
                        enumerate(self.network.hosts())
                        if machine_offset <= i + 1 < machine_offset + machine_count)

    def hosts(self) -> List[HostMeta]:
        return self._hosts



from ipaddress import IPv4Address, IPv4Network

from assertpy import assert_that

from netboot_config.network import HostGroup, HostMeta, Host


class TestHost(object):

    def setup_method(self):
        self.uut = Host('foo', IPv4Network('10.11.12.0/24'), IPv4Address('10.11.12.13'), 'base_image')

    def test_ipv4_address(self):
        result = self.uut.ipv4_address

        assert_that(result).is_equal_to('10.11.12.13')

    def test_reverse_pointer(self):
        result = self.uut.reverse_pointer

        assert_that(result).is_equal_to('13.12.11.10.in-addr.arpa')

    def test_hex_ipv4_address(self):
        result = self.uut.ipv4_address_hex

        assert_that(result).is_equal_to('0A0B0C0D')

    def test_host_name(self):
       result = self.uut.host_name

       assert_that(result).is_equal_to('foo13')


class TestNetwork(object):

    def setup_method(self):
        self.uut = HostGroup('foo', '10.11.12.0/24')

    def test_host_addresses(self):
        self.uut.add_hosts(11, 4, '<image_type>')

        hosts = self.uut.hosts()

        addresses = [host.ipv4_address for host in hosts]

        assert_that(addresses).contains_only('10.11.12.11', '10.11.12.12', '10.11.12.13', '10.11.12.14')

    def test_host_names(self):
        self.uut.add_hosts(11, 4, '<image_type>')

        hosts = self.uut.hosts()

        host_names = [host.host_name for host in hosts]

        assert_that(host_names).contains_only('foo11', 'foo12', 'foo13', 'foo14')

    def test_host_names_over_100(self):
        self.uut.add_hosts(111, 4, '<image_type>')

        hosts = self.uut.hosts()

        host_names = [host.host_name for host in hosts]

        assert_that(host_names).contains_only('foo11', 'foo12', 'foo13', 'foo14')



import os

from assertpy import assert_that

from netboot_config import Config


class TestConfig(object):

    def setup_method(self):
        dirname = os.path.dirname(__file__)
        filename = os.path.join(dirname, 'netboot-config.yml')
        self.uut = Config(filename)

    def test_netboot_ip(self):
        assert_that(self.uut.netboot_ip).is_equal_to("10.0.10.5")

    def test_hosts_names(self):
        hosts = self.uut.hosts
        assert_that([host.host_name for host in hosts]).contains("abc1110", "abc1111", "abc1120")

    def test_hosts_image_types(self):
        hosts = self.uut.hosts
        assert_that([host.image_type for host in hosts]).contains("type1", "type1", "type2")

    def test_hosts_image_config(self):
        hosts = self.uut.hosts
        assert_that([host.config for host in hosts]).contains((), (), (('src', 'tgt'),))

    def test_static_hosts_names(self):
        hosts = self.uut.static_hosts
        assert_that([host.host_name for host in hosts]).contains("abc1005", "abc1020", "abc1021", "abc1030", "abc1040")

    def test_static_hosts_image_types(self):
        hosts = self.uut.static_hosts
        assert_that([host.image_type for host in hosts]).contains(None, None, None, None, None)

    def test_static_hosts_image_aliases(self):
        hosts = self.uut.static_hosts
        assert_that([host.aliases for host in hosts]).contains((), (), (), ('foo',), ('bar', 'baz'))

    def test_all_hosts(self):
        assert_that(len(self.uut.all_hosts)).is_equal_to(8)

    def test_get_undefined_host_in_static_network(self):
        host = self.uut.get_host('10.0.10.99')
        assert_that(host.host_name).is_equal_to("abc1099")
        assert_that(host.aliases).is_empty()

    def test_get_defined_host_in_static_network(self):
        host = self.uut.get_host('10.0.10.30')
        assert_that(host.host_name).is_equal_to("abc1030")
        assert_that(host.aliases).contains("foo")

    def test_get_undefined_host_in_dynamic_network(self):
        host = self.uut.get_host('10.0.11.99')
        assert_that(host.host_name).is_equal_to("abc1199")
        assert_that(host.image_type).is_none()

    def test_get_defined_host_in_dynamic_network(self):
        host = self.uut.get_host('10.0.11.10')
        assert_that(host.host_name).is_equal_to("abc1110")
        assert_that(host.image_type).is_equal_to("type1")

    def test_host_outside_networks(self):
        host = self.uut.get_host('10.0.12.99')
        assert_that(host).is_none()

    def test_aliases(self):
        aliases = self.uut.aliases
        assert_that(aliases).contains(('foo', 'abc1030'), ('bar', 'abc1040'), ('baz', 'abc1040'))

    def test_render_hosts(self):
        result = self.uut.render_hosts()

        hosts = {}
        for line in result.split("\n"):
            parts = line.split("\t")
            if len(parts) == 2:
                hosts[parts[0]] = parts[1].split(" ")

        assert_that(hosts).contains_entry(
            {'127.0.0.1': ['localhost']},
            {'::1': ['localhost', 'ip6-localhost', 'ip6-loopback']},
            {'fe00::0': ['ip6-localnet']},
            {'ff00::0': ['ip6-mcastprefix']},
            {'ff02::1': ['ip6-allnodes']},
            {'ff02::2': ['ip6-allrouters']},
            {'ff02::3': ['ip6-allhosts']},

            {'10.0.11.10': ['abc1110']},
            {'10.0.11.11': ['abc1111']},
            {'10.0.11.20': ['abc1120']},
            {'10.0.10.5': ['abc1005', 'netboot']},
            {'10.0.10.20': ['abc1020']},
            {'10.0.10.21': ['abc1021']},
            {'10.0.10.30': ['abc1030', 'foo']},
            {'10.0.10.40': ['abc1040', 'bar', 'baz']})

    def test_render_hosts_returns_newline_at_end(self):
        result = self.uut.render_hosts()

        assert_that(result[-1]).is_equal_to('\n')

    def test_find_alias_by_name(self):
        result = self.uut.alias('foo')

        assert_that(result.host_name).is_equal_to("abc1030")

    def test_find_alias_by_name_no_result(self):
        result = self.uut.alias('qux')

        assert_that(result).is_none()



import ipaddress
from typing import Optional, List

import yaml

from .network import HostGroup, Host


class Config(object):

    def __init__(self, config_file: str):

        self._hosts = {}
        self._static_hosts = {}
        self.network_prefixes = {}
        self._aliases = {}

        with open(config_file, 'r') as stream:
            data_loaded = yaml.safe_load(stream)

            for host_group_entry in data_loaded['hostgroups']:
                host_group = self.create_host_group(host_group_entry)

                for host_entry in host_group_entry['hosts']:
                    self.map_entry(host_entry, host_group)

                self._hosts.update({host.ipv4_address: host for host in host_group.hosts()})

            for host_group_entry in data_loaded['hosts']:
                host_group = self.create_host_group(host_group_entry)

                for host_entry in host_group_entry['hosts']:
                    self.map_entry(host_entry, host_group)

                self._static_hosts.update({host.ipv4_address: host for host in host_group.hosts()})

            for host in self.all_hosts:
                for alias in host.aliases:
                    self._aliases[alias] = host

    def create_host_group(self, host_group_entry):
        prefix_ = host_group_entry['prefix']
        cidr_ = host_group_entry['cidr']
        self.network_prefixes[cidr_] = prefix_
        host_group = HostGroup(prefix_, cidr_)
        return host_group

    def map_entry(self, host_entry, host_group):
        offset_ = host_entry['offset']
        count_ = host_entry['count'] if 'count' in host_entry else 1

        if 'alias' in host_entry:
            aliases_ = (host_entry['alias'],)
        elif 'aliases' in host_entry:
            aliases_ = tuple(host_entry['aliases'])
        else:
            aliases_ = None

        if 'config' in host_entry:
            config_ = [(config_entry['source'], config_entry['target']) for config_entry in host_entry['config']]
        else:
            config_ = None

        image_type_ = host_entry['image_type'] if 'image_type' in host_entry else None

        host_group.add_hosts(offset_, count_, aliases_, image_type_, config_)

    @property
    def hosts(self) -> List[Host]:
        return list(self._hosts.values())

    @property
    def static_hosts(self) -> List[Host]:
        return list(self._static_hosts.values())

    @property
    def all_hosts(self) -> List[Host]:
        return self.hosts + self.static_hosts

    @property
    def aliases(self):
        aliases = []
        for alias, host in self._aliases.items():
            aliases.append((alias, host.host_name))
        return aliases

    @property
    def netboot_ip(self):
        return self.alias('netboot').ipv4_address

    def alias(self, alias):
        return self._aliases[alias] if alias in self._aliases else None

    def get_host(self, ip_address_string: str) -> Optional[Host]:
        if ip_address_string in self._static_hosts:
            return self._static_hosts[ip_address_string]

        if ip_address_string in self._hosts:
            return self._hosts[ip_address_string]

        ip_address = ipaddress.IPv4Address(ip_address_string)
        for cidr, prefix in self.network_prefixes.items():
            network = ipaddress.IPv4Network(cidr)
            if ip_address in network:
                return Host(prefix, network, ip_address)
        return None

    def render_hosts(self):
        lines = []

        lines.append("127.0.0.1\tlocalhost")
        lines.append("::1\tlocalhost ip6-localhost ip6-loopback")
        lines.append("fe00::0\tip6-localnet")
        lines.append("ff00::0\tip6-mcastprefix")
        lines.append("ff02::1\tip6-allnodes")
        lines.append("ff02::2\tip6-allrouters")
        lines.append("ff02::3\tip6-allhosts")

        lines.append("")

        for host in self.all_hosts:
            lines.append(host.entry)

        return "\n".join(lines) + "\n"



from assertpy import assert_that

from netboot_config.model import HostConfig


class TestHostConfig(object):

    def setup_method(self):
        self.uut = HostConfig()

    def test_renders_empty_config(self):
        result = self.uut.render()

        assert_that(result).is_none()

    def test_renders_single_entry(self):
        self.uut.add("foo", 'bar')

        result = self.uut.render()

        assert_that(result).is_equal_to('CONF="foo;bar;{netboot_ip};300"')

    def test_renders_multiple_entrie(self):
        self.uut.add("foo", 'bar')
        self.uut.add("baz", 'qux')

        result = self.uut.render()

        assert_that(result).is_equal_to('CONF="foo;bar;{netboot_ip};300,baz;qux;{netboot_ip};300"')



from .network import HostMeta


class HostConfig(object):

    def __init__(self):
        self.entries = []

    def add(self, source_path: str, target_path: str):
        self.entries += [(source_path, target_path)]

    def render(self) -> str:
        return "CONF=\"" + ",".join(
            ["{};{};{{netboot_ip}};300".format(entry[0], entry[1]) for entry in
             self.entries]) + "\"" if self.entries else None


class ConfigFile(object):

    def __init__(self, host: HostMeta, host_config: HostConfig, netboot_ip: str):
        self.host = host
        self.host_config = host_config
        self.netboot_ip = netboot_ip

    def write(self):
        with open('config.{}'.format(self.host.ipv4_address_hex), 'w') as kiwi_config_file:
            kiwi_config_file.write(
                "IMAGE=/dev/ram1;{};1.42.3;{};10096\n".format(self.host.image_type, self.netboot_ip))
            kiwi_config_file.write("UNIONFS_CONFIG=tmpfs,/dev/ram1,overlay\n")
            host_config_data = self.host_config.render()
            if host_config_data:
                kiwi_config_file.write("{}\n".format(host_config_data.format(netboot_ip=self.netboot_ip)))



# coding=utf-8
import glob

import setuptools

setuptools.setup(
    name="netboot-config",
    version="0.1.28",
    author="Andreas Würl",
    author_email="andreas.wuerl@uniscon.com",
    description="Generator for KIWI based netboot config files",
    long_description_content_type="text/markdown",
    packages=["netboot_config"],
    scripts=["netboot-config"],
    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        "Programming Language :: Python :: 3.5",

        "License :: OSI Approved :: Apache Software License",

        "Operating System :: POSIX :: Linux",
    ],
)
